# inspired by example at
# https://github.com/picotech/picosdk-python-wrappers/blob/master/usbtc08Examples/tc08StreamingModeExample.py
#
# API docs
# https://www.picotech.com/download/manuals/usb-tc08-thermocouple-data-logger-programmers-guide.pdf

import ctypes
import signal
import time
from threading import Timer
from picosdk.usbtc08 import usbtc08 as tc08
from influxdb_bridge import InfluxDBBridge

# how often does the TC08 driver collect data
POLL_INTERVAL_MS = 5000
# how often does Monitor check for new data from the TC08 driver
LOG_INTERVAL_SEC = 5

DEBUG = False

ERROR_CODES = [
    'USBTC08_ERROR_OK',
    'USBTC08_ERROR_OS_NOT_SUPPORTED',
    'USBTC08_ERROR_NO_CHANNELS_SET',
    'USBTC08_ERROR_INVALID_PARAMETER',
    'USBTC08_ERROR_VARIANT_NOT_SUPPORTED',
    'USBTC08_ERROR_INCORRECT_MODE',
    'USBTC08_ERROR_ENUMERATION_INCOMPLETE',
    'USBTC08_ERROR_NOT_RESPONDING',
    'USBTC08_ERROR_FW_FAIL',
    'USBTC08_ERROR_CONFIG_FAIL',
    'USBTC08_ERROR_NOT_FOUND',
    'USBTC08_ERROR_THREAD_FAIL',
    'USBTC08_ERROR_PIPE_INFO_FAIL',
    'USBTC08_ERROR_NOT_CALIBRATED',
    'USBTC08_ERROR_PICOPP_TOO_OLD',
    'USBTC08_ERROR_COMMUNICATION'
]


# Channel 0 is ambient interface board temp
# Channel 8 starts at furthest SE side and progresses counterclockwise
LOCS = ['Ambient', 'SW S', 'SW N', 'NW S',
        'NW N', 'NE N', 'NE S', 'SE N', 'SE S']


def dprint(arg):
    if DEBUG:
        print(arg)


class Monitor(Timer):
    def __init__(self, interval):
        # Timer constructer expects a callback function, but we don't care
        # about that, so just pass None
        super(Monitor, self).__init__(interval, None)
        self.chandle = ctypes.c_int16()
        self.status = {}
        # self.timer = RepeatTimer(POLL_INTERVAL_MS, self.injest_data)
        self.idb = InfluxDBBridge()
        dprint('Monitor initialized')

    # override Timer's run method to loop on our injest_data function
    # wait will block while loop for timeout interval
    # calling timer.cancel() sets the finish flag and exits the loop
    def run(self):
        dprint('in timer.run method')
        while not self.finished.wait(timeout=self.interval):
            self.injest_data()

    # helper to check return codes
    def check_return(self, rc):
        if rc <= 0:
            dprint(f'last return code: {rc}')
            # tc08 functions seem to return 0 or -1 on failure, in which case
            # we can querry get_last_error for a more detailed error code
            error_code = tc08.usb_tc08_get_last_error(self.chandle)
            if error_code == -1:
                raise Exception(
                    f'Error retrieving last error (oh the irony) (return code: {rc})')
            if error_code >= 0 and error_code < len(ERROR_CODES):
                raise Exception(
                    f'Error {error_code}: {ERROR_CODES[error_code]}')
            else:
                raise Exception(f'Unknown error code: {error_code}')

    # open and configure all channels for K-type thermistors
    def setup(self):
        dprint('begin setup')
        # open unit
        dprint('opening unit')
        self.status["open_unit"] = tc08.usb_tc08_open_unit()
        self.check_return(self.status["open_unit"])
        self.chandle = self.status["open_unit"]
        dprint('tc08_open_unit complete')

        # set mains rejection to 50 Hz
        dprint('setting mains rejection')
        self.status["set_mains"] = tc08.usb_tc08_set_mains(self.chandle, 0)
        self.check_return(self.status["set_mains"])
        dprint('tc08_set_mains complete')

        # set up all channels (channel 0 is ambient "CJC" or "cold junction")
        for channel in range(0, 9):
            self.setup_channel(channel)

        # get minimum sampling interval in ms
        self.status["get_minimum_interval_ms"] = tc08.usb_tc08_get_minimum_interval_ms(
            self.chandle)
        self.check_return(self.status["get_minimum_interval_ms"])

        # open connection to influxdb
        self.idb.connect()

        dprint('end setup')

    # helper to set up each channel
    def setup_channel(self, channel):
        dprint(f'setup_channel {channel}')
        # therocouples types and int8 equivalent
        # B=66 , E=69 , J=74 , K=75 , N=78 , R=82 , S=83 , T=84 , ' '=32 , X=88
        typeK = ctypes.c_int8(75)
        rc = tc08.usb_tc08_set_channel(self.chandle, channel, 75)
        self.check_return(rc)
        dprint(f'setup_channel({channel}) complete')

    def tc08_run(self):
        # set tc-08 running with POLL_INTERVAL_MS, unless it is below the minimum possible interval
        poll_interval = max(
            POLL_INTERVAL_MS, self.status["get_minimum_interval_ms"])
        dprint(f'tc08_run begin with polling interval {poll_interval} ms')
        self.status["run"] = tc08.usb_tc08_run(
            self.chandle,
            poll_interval
        )
        self.check_return(self.status["run"])
        self.start_time_sec = time.time()
        dprint('tc08 running')

    def tc08_stop(self):
        # stop unit
        self.status["stop"] = tc08.usb_tc08_stop(self.chandle)
        self.check_return(self.status["stop"])
        dprint('tc08 stopped')

    def tc08_close(self):
        # close unit
        self.status["close_unit"] = tc08.usb_tc08_close_unit(self.chandle)
        self.check_return(self.status["close_unit"])
        dprint('tc08 closed')

    # tc08 reports times in millis since run started
    def convert_time(self, relative_time_ms):
        return int((self.start_time_sec * 1000) + relative_time_ms)

    def injest_data(self, buffer_size=8):
        dprint('entering injest_data')
        temperature_buffer = (ctypes.c_float * buffer_size)()
        times_ms_buffer = (ctypes.c_int32 * buffer_size)()
        overflow = ctypes.c_int16()
        # in streaming mode, data is buffered by the driver and we must
        # retrieve it for individual channels
        for channel in range(0, 9):
            # load in up to buffer_size buffered readings
            num_readings = tc08.usb_tc08_get_temp(
                self.chandle,
                ctypes.byref(temperature_buffer),
                ctypes.byref(times_ms_buffer),
                buffer_size,
                ctypes.byref(overflow),
                channel,
                tc08.USBTC08_UNITS['USBTC08_UNITS_CENTIGRADE'],
                0
            )
            # if num_readings is -1, there was an error
            if num_readings < 0:
                self.check_return(num_readings)
            elif num_readings == 0:
                # no data, nothing to do
                dprint('no data read')
            else:
                for i in range(0, num_readings):
                    temperature = temperature_buffer[i]
                    time_ms = times_ms_buffer[i]
                    # insert data into db
                    self.idb.commit_point(
                        channel,
                        LOCS[channel],
                        temperature,
                        self.convert_time(time_ms),
                        overflow.value)
                    # self.idb.print_point(channel, temperature, time_ms)

    def start_monitor(self):
        self.setup()
        self.tc08_run()
        self.start()

    def cancelISR(self, *args, **kwargs):
        dprint('cancelISR called')
        self.cancel()
        self.tc08_stop()
        self.tc08_close()


if __name__ == '__main__':
    monitor = Monitor(LOG_INTERVAL_SEC)
    monitor.start_monitor()
    dprint('monitor running')
    # SIGINT isn't getting caught even with this, I think the ctypes calls
    # made by the picotech tc08 library are disrupting python's signal handling
    # I don't have a workaround yet (maybe re-register the handler after every
    # return from usb_tc08_get_temp?)
    signal.signal(signal.SIGINT, monitor.cancelISR)
